/* James Houghton
 * 06 Dec 2018
 * FTP Server: Connection establishment
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include "server.h"

static int socketfd = -1;

// Start the server
int start(unsigned short port) {

    // Open the main control socket
    struct sockaddr_in sa;
    socketfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (socketfd == -1) {
        perror("cannot create socket");
        exit(EXIT_FAILURE);
    }

    // Allow re-use of the port
    int tr = 1;
    setsockopt(socketfd,SOL_SOCKET,SO_REUSEADDR,&tr,sizeof(int));

    // Set the host mask and port
    memset(&sa, 0, sizeof sa);
    sa.sin_family = AF_INET;
    sa.sin_port = htons(port);
    inet_pton(AF_INET, "127.0.0.1", &sa.sin_addr.s_addr);
    // sa.sin_addr.s_addr = htonl(INADDR_ANY);

    // Bind to port
    if(bind(socketfd, (struct sockaddr *)&sa, sizeof sa) == -1) {
        perror("bind failed");
        close(socketfd);
        exit(EXIT_FAILURE);
    }

    // Listen for connections
    if(listen(socketfd, 10) == -1) {
        perror("listen failed");
        close(socketfd);
        exit(EXIT_FAILURE);
    }

    /* Main loop */
    for(;;) {
        puts("Waiting for a connection");

        // Keep track of FTP connection, set to defaults
        struct ftp_conn_t conn;
        conn.ft_type          = ASCII;
        conn.struct_type      = FLE;
        conn.mode             = STREAM;
        conn.dtp_type         = ACTIVE;
        conn.client_addr_size = sizeof(conn.client_addr);
        conn.data             = -1;
        conn.cwd_length       = 1;
        conn.cwd[0] = '/';
        conn.cwd[1] = '\0';

        int connectfd = accept(socketfd, (struct sockaddr *)&conn.client_addr,
                &conn.client_addr_size);

        // FTP connection established

        if(connectfd < 0) {
            perror("accept failed");
            close(socketfd);
            continue;
        }

        conn.sock = connectfd;

        char ip[INET_ADDRSTRLEN];
        inet_ntop(AF_INET, &conn.client_addr.sin_addr, ip, conn.client_addr_size);
        printf("Established connection to %s:%hu\n",
                ip, ntohs(conn.client_addr.sin_port));

        // Set default data channel to client connection address, per spec
        // Subsequent calls to PORT will may this
        memcpy(&conn.data_addr, &conn.client_addr, sizeof(conn.client_addr));

        // Fork off a worker
        // This is to allow for multiple simultaneous connections
        pid_t pid = fork();
        if(pid < 0) {
            perror("error spawning worker");
        }
        else if(pid == 0) {
            // Start the command dispatcher
            interp(&conn);

            // Shutdown the connection
            if(shutdown(connectfd, SHUT_RDWR) == -1) {
                perror("shutdown failed");
            }

            puts("Closing connection");

            // Close the data connection if it is open
            if(conn.data >= 0) {
                puts("Closing data connection");
                close(conn.data);
            }
            close(connectfd);
            puts("Worker exiting");
            exit(EXIT_SUCCESS);
        } else {
            printf("Spawned worker: %d\n", pid);
            close(connectfd);
        }
    }

    close(socketfd);
    return EXIT_SUCCESS;
}

void sighandle(int sig) {
    close(socketfd);
    exit(EXIT_SUCCESS);
}

int main(int argc, char **argv) {
    if(argc < 2) {
        fputs("Please specify the port number you wish to use\n", stderr);
        return 1;
    }

    int port = atoi(argv[1]);
    if(port <= 0 || port >> 16 > 0) {
        fputs("Invalid port specified\n", stderr);
        return 1;
    }

    signal(SIGINT, &sighandle);

    return start((unsigned short)port);
}
