/* James Houghton
 * 12/12/18
 * FTP Connection Functions and Types
 */

#ifndef INTERPRETER_H
#define INTERPRETER_H

#define FTP_SUCCESS 0
#define FTP_QUIT 1
#define FTP_ERROR 2
#define FTP_NOENT 3
#define FTP_PREEXISTS 4

#define COMMAND_LENGTH 4096
#define PATH_LENGTH 4096

#include <netinet/in.h>

enum file_trans_t { ASCII, IMAGE };
enum struct_t     { FLE, RECORD, PAGE };
enum mode_t       { STREAM, BLOCK, COMPRESSED };
enum dtp_t        { ACTIVE, PASSIVE };

struct ftp_conn_t {
    int sock;                       // Control socket

    enum file_trans_t ft_type;      // File transport type
    enum struct_t struct_type;      // File structure (always FLE, file)
    enum mode_t mode;               // Transmission mode (always STREAM)
    enum dtp_t  dtp_type;           // Data transfer process (always PASSIVE)

    struct sockaddr_in client_addr; // Client control address
    socklen_t client_addr_size;

    struct sockaddr_in data_addr;   // Data socket
    int data;

    char cwd[4096];                 // The working directory (as seen by the client)
    size_t cwd_length;              // Real CWD length
};

// See implementation for comments
void interp(struct ftp_conn_t *);

void ftp_send(struct ftp_conn_t *, const char *);
int open_data_conn(struct ftp_conn_t *);
int close_data_conn(struct ftp_conn_t *);

int send_error(struct ftp_conn_t *, int);

#endif
